/*
 * pbm2l7k.c -- convert monochromatic PBMraw file into Lexmark 7000
 *              printing escape sequences
 *              Usage: ./pbm2l7k < file.pbm > /dev/lp1
 *              
 * (C) 1999 Henryk Paluch, released under GNU GPL license, version 2
 * 
 * $Id: pbm2l7k.c,v 1.29 1999/05/15 14:50:40 henryk Exp $
 */

#include<stdlib.h>
#include<stdio.h>
#include<errno.h>
#include<string.h>
#include<unistd.h> /* getopt() */
#include<string.h>
#include<errno.h>
#include<syslog.h>

#include"lxmisc.h"

/* this will be model dependent section */
enum LX_MODELS { LXM_7000=0,LXM_5700,LXM_LAST};
int lx_model=LXM_7000; /* current selected Lexmark printer model */

struct lx_smodel{
   char printer_name[16];
   char prologoue_name[64];
};

struct lx_smodel lx[]={{"Lexmark 7000","pr7000.prn"},
                            {"Lexmark 5700","pr5700.prn"}};

#ifndef MIN
#define MIN(x,y) ( (x) < (y) ? (x) : (y) )
#endif



#define DIV8(x) ( (x) >> 3 )
#define MOD8(x) ( (x) & 0x7 )

#define BLINES (lines)
			    /* offset for B&W printing */
#define CLINES (lines)
			    /* offset for Cyan plate */
#define MLINES (lines+(1*pheight*width))
			    /* offset for Magenta plate */
#define YLINES (lines+(2*pheight*width))
			    /* offset for Yellow plate */
#define KLINES (lines+(3*pheight*width))
			    /* offset for additional K (black) plate */

int height;                 /* height of PBM bitmap (lines e.g., pixels) */
int width;                  /* width of line bitmap + LR_MARGINS 
			     * (in bytes)
			     */
int bwidth;                 /* real bitmap width (bytes) */
int lr_shift8;              /* LEFT-RIGHT shift (bytes) */
int pheight;		    /* height of one color plate */
                            /* aligned to VERTSIZE or st similar... */
int psize;		    /* plate size */
unsigned char *lines=NULL; /* this is pixel buffer for one pass */
int plates;              /* number of color plates:
			    * 1 = traditional Black & White
			    * 3 = CMY
			    * 4 = CMYK
			    */
int passheight;          /* 192 for B&W or 192/3=64 for CMY colour */
int mayeof;

/*printing constants*/
const int  TOP_MARGIN=0x100;
const int  LEFT_MARGIN=50;
const int  VERT_SHIFT=0x180;  /* paper shift betwen printed lines */
const int  LR_SHIFT=16; /* distance betwen two series of Left & Right
			   inkjets */

const int VCOL_SPACE=12; /* vertical space between two colors */


/* number of vertical lines printed in one pass 
 * Lexmark 7000 has (probably) 2 series of 12 * 8 = 96 InkJets
 * e.g. total 2 * 96 = 192 is height (in pixels) printed in one pass
 *
 * don't change this, unless you know what you are doing!
 */
const int VERTSIZE=192;

/* offsets to print line sequence (defined in outbuf)
 */
const int IDX_SEQLEN=5;
const int IDX_PACKETS=13;
const int IDX_HSTART=15;
const int IDX_HEND=17;
const int IDX_DATA=26;
const int IDX_CARTRIDGE=10;

#define OUT_BUF_SIZE 65535
/* most important print packet
 * see lexmarkprotocol.txt for more information */
/*          length of complete sequence ---  vvvvvvvvv */
char outbuf[OUT_BUF_SIZE]={0x1B,0x2A,0x04,0x00,0x00,0xFF,0xFF,
   /* number of packets ----     vvvvvvvvv */ 
   0x00,0x02,0x01,0x01,0x1A,0x11,0xFF,0xFF,
   /* horiz start, horiz end: packets = (horiz end - horiz start) +1 */
   0xFF,0xFF,0xFF,0xFF,0x00,0x00,0x22,0x33,0x44,0x55,0x01};

/* fast_cat() - just emulates cat
 *              writes 'filename' to stdout
 *              Used to print captured prologue 'prolog.prn' to printer
 */
void fast_cat(char *filename)
{
  FILE *f;
  char buf[1024];
  int n;

  if ( (f=fopen(filename,"r"))==NULL )
     lxmsg(LXV_FATAL,"fopen '%s': %s\n",filename,strerror(errno));

  while (  (n=fread(buf,1,1024,f))>0)
    if ( fwrite(buf,1,n,stdout)!=n)
       lxmsg(LXV_FATAL,"fast_cat: error while copying\n");
 fclose(f);
}

/*  parse_pbm_header() - decodes header of PBM bitmap
 *        sets global vars: height (pixels), 
 *                          bwidth (real bitamp width - bytes)
 *                          width  (memory bitmap width - bytes)
 *        PBMraw is simple monochromatic bitmap. It consists
 *        of text header and raw data bytes.
 *        Example: P4
 *                 # comment lines
 *                 width height
 *                 raw bytes
 *                 
 *                 width & height is in pixels, lines are byte aligned
 *                 0x80 is most left bit, 0x1 most right
 */
int parse_pbm_header(FILE *f,int *width, int *height,int *plates)
{
   char buf[1024];
   int dontscan=0;
   
   if (fgets(buf,1023,f)==0  || strncmp(buf,"P4",2)!=0)
   {
      if (mayeof)
	 return 0;
      else
	 lxmsg(LXV_FATAL,"Illegal pbmraw header");
   }
   *plates=1; /* assume B&W PBM RAW bitmap on input */
   do{
      fgets(buf,1023,f); /* skip comment line(s) */
      if ( strncmp(buf,"#PI",3)==0 )  /* interlaced colors CMY(K) lines? */
      {
	 *plates=(int)(buf[3]-'0');
	 lxmsg(LXV_STD,"Detected %d colors interlaced PBM bitmap\n",*plates);
	 sscanf(buf+4,"%d %d",width,height);
	 dontscan=1;
      }
   }while(buf[0]=='#');

   if (!dontscan)
      sscanf(buf,"%d %d\n",width,height);
   if ( (*width) <=0 || (*height) <=0)
      lxmsg(LXV_FATAL,"Illegal width (%d) or height (%d)\n",*width,*height);
   return 1;
}

/* paper_shift() -- shift paper in printer -- units are unknown :-)
 */
void paper_shift(int offset)
{
   unsigned char buf[5]={0x1b,0x2a,0x3,0x0,0x0};
   buf[3]=(unsigned char)(offset >> 8);
   buf[4]=(unsigned char)(offset & 0xFF);
   if (fwrite(buf,1,5,stdout)!=5)
      lxmsg(LXV_FATAL,"paper_shift: short write\n");
}

/* paper_eject() -- self explaining
 */
void paper_eject(void)
{
   unsigned char buf[4]={0x1b,0x2a,0x7,0x65};
   if (fwrite(buf,1,4,stdout)!=4)
      lxmsg(LXV_FATAL,"paper_eject: short write\n");
}

/* read_lines() -- read 'readcount' data lines from PBM bitmap
 *                 to 'lines' -- bitmap in memory
 */

int read_lines(FILE *f,int readcount)
{
   int i,c;

   if (lines==NULL)
      lines=(char*)lxmalloc(width*VERTSIZE);
   /* zero the buffer */
   memset(lines,(char)0,width*VERTSIZE);
   for(i=0;i<readcount;i++)
   {
      c=fread(lines+lr_shift8+width*i,1,bwidth,f);
      if (c==-1)
	 lxmsg(LXV_FATAL,"read_lines: fread: %s\n",strerror(errno));
      if (c<bwidth)
	 return c;
   }
   return c;
}

void read_plate(FILE *f,unsigned char *buf)
{
   int i,c;
   memset(buf,(char)0,psize);
   for(i=0;i<height;i++)
   {
      c=fread(buf+lr_shift8+width*i,1,bwidth,f);
      if (c==-1)
	 lxmsg(LXV_FATAL,"read_plate: fread: %s\n",strerror(errno));
      if (c<bwidth)
	 lxmsg(LXV_FATAL,"read_plate: short read %d of %d %s\n",
	       strerror(errno),c,bwidth);
   }
}  

 
/* read_color_plates() for one pass
 * f - file to read
 * buf - buffer
 * horizontal offset (bytes)
 * width - to be read (bytes)
 * real buffer width (bytes)
 * buffer height (bytes)
 * xheight (lines to read - may be < height at the end of bitmap)
 * pheight - pass height (now 64 for all the time)
 */
void read_color_plates(FILE *f,unsigned char *buf, int ofsx, int bwidth,
       int width, int height,int xheight,int pheight,int plates)
{
   int i,j,c;

   /* 1st - old data to pheight up */
   for(j=0;j<plates;j++)
   {
      int plofs=height*width*j;
      memmove(buf+plofs,buf+(plofs+pheight*width),(height-pheight)*width);
   }

   memset(buf+((height-pheight)*width),(char)0,width*pheight);
   for(i=0;i<xheight;i++)
   {
      for(j=0;j<plates;j++)
      {
	 c=fread(buf+((height-pheight)*width+j*height*width+(ofsx+width*i)),1,bwidth,f);
	 if (c==-1)
	    lxmsg(LXV_FATAL,"read_color_plates: %s\n",strerror(errno));
	 if (c<bwidth)
	    lxmsg(LXV_FATAL,"read_color_plates: short read\n");
      }
   }
}  


 
/* read_mono_plate() for one pass
 * f - file to read
 * buf - buffer
 * horizontal offset (bytes)
 * width - to be read (bytes)
 * real buffer width (bytes)
 * buffer height (bytes)
 * xheight (lines to read - may be < height at the end of bitmap)
 */


 void read_mono_plate(FILE *f,unsigned char *buf, int ofsx, int bwidth,
       int width, int height,int xheight)
{
   int i,c;
   int clines=height-xheight;
   lxmsg(LXV_DEBUG,"%d lines to clear before read\n",clines);
   if (clines>0)
      memset(buf+(xheight*width),(char)0,width*clines);
   for(i=0;i<xheight;i++)
   {
      c=fread(buf+(ofsx+width*i),1,bwidth,f);
      if (c==-1)
	 lxmsg(LXV_FATAL,"read_mono_plate: fread: %s\n",strerror(errno));
      if (c<bwidth)
	 lxmsg(LXV_FATAL,"read_mono_plate: short read %d of %d\n",
	       c,bwidth);
   }
}  
   

  

/* are_lines_empty() -- return true, if all VERTSIZE 'lines' are
 *                      empty => nothing to print => only paper shift
 */
int are_lines_empty(unsigned char *buf)
{
   int i;
   for(i=0;i<width*passheight;i++)
      if (buf[i]!=0)
	 return 0;
   return 1;
}

/*  is_col_empty() -- true, if column 'col' (in pixels) is empty
 */
int is_col_empty(int col,unsigned char *buf) /* col is now in pixels! */
{
   unsigned char mask[8]={128,64,32,16,8,4,2,1};
   int i;
   unsigned char *bofs, *bofs2;
   unsigned char mmask, mmask2;
   mmask=mask[MOD8(col)];
   mmask2=mask[MOD8(col+LR_SHIFT)];
   bofs=buf+DIV8(col);
   bofs2=buf+width+DIV8(col+LR_SHIFT);

   for(i=0;i<passheight/2;i++)
   {
      if ( ( bofs[0] & mmask)!=0 ) /* left series */
	 return 0;
/*      if (col+LR_SHIFT<width*8) */
        if ( (bofs2[0] & mmask2)!=0 ) /* right series */
          return 0;
      bofs+=width*2; /* aaddres + 2 lines */
      bofs2+=width*2;
   }
   return 1;
}

/*   find_horiz_ofsets() -- returns horizontal offsets for printing
 *   (e.g. left = 1st pixel on the left, right= last pixel on the row
 *   
 */
void find_horiz_ofsets(int *left,int *right,unsigned char *buf)
{
   int i;
   *left=0;
   *right=width*8-LR_SHIFT; /* left & right is in pixels (bits) now! */
   for(i=0;i<width*8;i++)
   {
      if (!is_col_empty(i,buf))
      {
	 *left=i;
	 break;
      }
   }
   for(i=width*8-LR_SHIFT-1;i>=0;i--)
      if (!is_col_empty(i,buf))
      {
	 *right=i+1;
	 break;
      }
}

/*   print_cols() -- most important -- print complete graphics line
 *                   in one pass
 *                   now 600x600dpi, but note, that we must handle
 *                   two series of distant ink jets (distance is
 *                   in (LR_SHIFT)
 */
int print_cols(int left,int right,int vstart, int vend,unsigned char *buf,
      int color)
{
   unsigned char mask[8]={128,64,32,16,8,4,2,1};
   unsigned int masq[12]={2048,1024,512,256,128,64,32,16,8,4,2,1};
   int l8,r8,packets;
   char *p=outbuf+IDX_DATA;
   int clen;
   int i,j,k,m;
   char *tbits;
   int bits;
   int byte;
   int word;
   unsigned char vbuf[VERTSIZE/2]; // Always corresponds to VERTSIZE !!!
   unsigned char *vp,*vp2;
   int smask,smask2;
   
   l8=left+LEFT_MARGIN;
   r8=right+LEFT_MARGIN;

   packets=r8-l8+1;
   
   if (lx_model==LXM_5700)
      outbuf[IDX_PACKETS-1]=(unsigned char) (color ? 0x11 : 0x1);
              /* Why is 5700 so different, ugh? */

   outbuf[IDX_PACKETS]  =(unsigned char)(packets >> 8);
   outbuf[IDX_PACKETS+1]=(unsigned char)(packets & 0xFF);
   outbuf[IDX_HSTART]   =(unsigned char)(l8 >> 8);
   outbuf[IDX_HSTART+1] =(unsigned char)(l8 & 0xFF);
   outbuf[IDX_HEND]     =(unsigned char)(r8 >> 8);
   outbuf[IDX_HEND+1]   =(unsigned char)(r8 & 0xFF);
   
   outbuf[IDX_CARTRIDGE]= (unsigned char)( color ? 0 : 1 );
   /* fill columns */
   for(i=left;i<right;i++)
   {
      if ( (p-outbuf)>=OUT_BUF_SIZE-26 ) /* 12*2+2 */
      {
	 return -1;
      }
	 if (is_col_empty(i,buf))
	 {
	    *(p++)=0x3F;
	    *(p++)=0xFF; /* whole column is empty */
	 }
	 else
	 {
	    *(p++)=0x3F;
	    tbits=p;
	    *tbits=0; /* here will be nice bitmap */
	    p++;
	    bits=0;
	    
	    /* clear buffer of vertical points */
	    memset(vbuf,0,VERTSIZE/2);
	    vp=buf+DIV8(i);
	          /* vp points to current column for Left Ink */
	    vp2=buf+width+DIV8(i+LR_SHIFT);
	          /* vp2 for right Ink jets */
	    smask=mask[MOD8(i)];
	    smask2=mask[MOD8(i+LR_SHIFT)];
	    for(k=vstart;k<vend;k++)
	    {
	       if ( ( vp[0] & smask) != 0)
		  vbuf[k] |= 1;
	       if (i+LR_SHIFT<width*8)
	       {
	          if ( (vp2[0] & smask2) !=0 )
		     vbuf[k] |=2;
		     vp2+=width*2;
	       }
	       vp+=width*2;
	    }

	    /* every packet contains 12 info bits
	     * 1 = 8x left white, 8x right white
	     * 0 = 2 data bytes follow (8 bits left, 8 bits right)
	     */

	    for(k=0;k<12;k++)
	    {
	       for(m=0;m<8;m++)
		  if ( vbuf[k*8+m]!=0 )
		     break;
	       if(m==8)
	       { /* 8vertical bits are empty */
		 bits= (bits<<1);
           	bits+=1; 
	       }
	       else
	       {
		  bits= (bits<<1);
		  byte=0;
		  for(m=0;m<8;m++)
		  {
		     byte= (byte << 1);
		     if ( (vbuf[k*8+m] & 1) != 0 )
			byte+=1;
		     byte= (byte << 1);
		     if ( (vbuf[k*8+m] & 2) != 0 )
			byte+=1; 
		  }
		  *(p++)= ((byte>>8) & 0xFF);
		  *(p++)= ( byte & 0xFF);
	       }
	    }
	    word=0;
	    for(m=0;m<12;m++)
	    {
	       word= (word << 1);
	       if ( bits & masq[11-m])
		  word +=1;
	    }
	    *tbits=(unsigned char)( (word) & 0xFF);
	    *(tbits-1)=(unsigned char)( ((word>>8) & 0xf) | 0x30 );
	 }
   }
   
   
   /* ------------ */
   clen=p-outbuf; 
   outbuf[IDX_SEQLEN]  =(unsigned char)(clen >> 8);
   outbuf[IDX_SEQLEN+1]=(unsigned char)(clen & 0xFF);
   if ( fwrite(outbuf,1,clen,stdout)!=clen)
      lxmsg(LXV_FATAL,"print_cols: short write\n");
   return 0;
}

static void usage(void)
{
   int i;
   fprintf(stderr,"Lexmark 7000 printer driver for Linux\n\n");
   fprintf(stderr,"Usage: pbm2l7k [ -h ] [ -m mode ] [ -v num ]"
	          " < input.pbmraw > output.prn\n");
   fprintf(stderr,"       -h\tthis help text\n");
   fprintf(stderr,"       -m mode\tspecifies printer driver:\n");
   for(i=0;i<LXM_LAST;i++)
      fprintf(stderr,"       -m %d\t%s\n",i,lx[i].printer_name);
   fprintf(stderr,"\n default mode is: [%d] %s\n",
	 lx_model,lx[lx_model].printer_name);
   fprintf(stderr,"       -v num\t verbosity 0=errors, 1=quiet, "
	          "2=standard, 3=debug\n"
		  "             \t default: %d\n",lxverbosity);
   fprintf(stderr,"\nColor printing is turned on, when interlaced"
	          " dithered CMY(K) bitmaps\nare produced by"
		  " pnmra2cmyk - mode is detected reading \"#PI? w h\"\n"
		  "where: ? = 3 (CMY) or 4 (CMYK)\n" 
		  "       w = width of one dithered color bitmap\n"
		  "       h = height of one dithered color bitmap\n");
   exit(-1);
}

int main(int argc, char *argv[])
{
   int i,linepairs;
   int left,right;
   int elinescount=0;
   int rlines;
   int c;
   int pmode;
   int pagecount=0;
   int twopasscount=0;
   int lremain;

   while( (c=getopt(argc, argv, ":m:hv:"))!=-1 )
      switch(c){
	 case 'h':
	    usage();
	    break;
	 case 'm':
	    pmode=atoi(optarg);
	    if (pmode<=0 || pmode>=LXM_LAST)
	    {
	       fprintf(stderr,"Invalid mode '%d' specified,"
		     " must be between <0,%d>\n", pmode, LXM_LAST-1);
	       exit(-1);
	    }
	    lx_model=pmode;
	    break;
	 case 'v':
	    lxverbosity=atoi(optarg);
	    break;
	 default:
	    usage();
      }


   openlog("pbm2l7k", LOG_PID|LOG_PERROR, LOG_LPR);
   lxusesyslog=1;

   lxmsg(LXV_STD,
	 "$Id: pbm2l7k.c,v 1.29 1999/05/15 14:50:40 henryk Exp $ started\n");
   lxmsg(LXV_STD,"Using driver: '%s'\n",lx[lx_model].printer_name);
   lxmsg(LXV_DEBUG,"Using driver for '%s', prologue file '%s'\n",
	 lx[lx_model].printer_name,
	 lx[lx_model].prologoue_name);
   /* write out printer prolog */
   //   fast_cat(lx[lx_model].prologoue_name);

   /* process raw PBM bitmap(s) */
   while(parse_pbm_header(stdin,&width,&height,&plates))
   {
      int plines;
      int hofs; /* highest vertical offset in buffer (meaning cartridge) */
      char *prmode[]={"Black & White","","CMY color", "CMYK color"};
      mayeof=1;
      bwidth= (width+7) /8; /* round width to whole byte */
      lxmsg(LXV_DEBUG,"bwidth (bytes): %d (%d pixels), height (lines): %d\n",
	    bwidth,bwidth*8, height);
      lr_shift8=(LR_SHIFT+8)/8;
      width=bwidth+2*lr_shift8;
      lxmsg(LXV_DEBUG,"width: %d (%d pixels)\n",width,width*8);
      lremain=height;

      lxmsg(LXV_STD,"printing page %d in %s mode\n",pagecount+1,
	    prmode[plates-1]);

      if (plates==1)
      {
	 passheight=192;
	 pheight=passheight;
	 hofs=0;
      }
      else
      {
	 passheight=192/3;
	 hofs=2*64+2*VCOL_SPACE;
	 pheight=passheight+hofs+2*64; // add space for more colors
      }

      psize=width*pheight; /* size of one plate */

      elinescount=TOP_MARGIN; /* initial paper shift */

      lxmsg(LXV_DEBUG,"One pass contains %d lines\n", passheight);

      lxmsg(LXV_DEBUG,"Plate height is %d pixels\n",pheight);
      lxmsg(LXV_DEBUG,"Plate size is %d bytes\n",psize);
      lxmsg(LXV_DEBUG,"Allocated %d plates of %d bytes\n",plates,plates*psize);
      lines=(unsigned char*)lxmalloc(psize*plates);
      memset(lines,(char)0,psize*plates); /* needed at least for color printing */

      while(lremain>0)
      {
	 plines=MIN(lremain,passheight);
	 lxmsg(LXV_DEBUG,"Will read %d lines for this pass\n",plines);
	 lxmsg(LXV_DEBUG,"%d lines to finish\n",lremain);

	 if (plates==1)
	 { /* Black & White printing mode */
	    read_mono_plate(stdin,lines,lr_shift8,
		  bwidth, /*bitmap width bytes read per line*/
		  width,  /*real buffer width>bwidth (bytes)*/
		  pheight, /* buffer height */
		  plines
			   );
	    if (!are_lines_empty(lines))
	    {
	       if (elinescount!=0)
	       {
		  paper_shift(elinescount);
		  elinescount=0;
	       }
	       find_horiz_ofsets(&left,&right,lines);
	       lxmsg(LXV_DEBUG,"offsets (pixels): left: %d, right %d\n",left,right);
	       if (print_cols(left,right,0,VERTSIZE/2,lines,0)==-1)
	       {
		  print_cols(left,right,0,VERTSIZE/4,lines,0);
		  print_cols(left,right,VERTSIZE/4,VERTSIZE/2,
			lines+192/2*width,0);
		  twopasscount++;
	       }

	    }
	    else
	    {
	       lxmsg(LXV_DEBUG,"192 empty lines\n");
	    }
	    elinescount+=VERT_SHIFT;
	 }
	 if (plates>=3)
	 {
	    /* color printing (now CMY mode only) */
	    /* handle yellow now */
	    unsigned char *ybuf;
	    unsigned char *mbuf;
	    unsigned char *cbuf;
	    read_color_plates(stdin,lines,lr_shift8,
		  bwidth, /*bitmap width bytes read per line*/
		  width,  /*real buffer width>bwidth (bytes)*/
		  pheight, /* buffer height */
		  plines, /* lines to be read <=passheight */
		  passheight, /* pass height (64) for color printing */
		  plates  /* number of color plates */
			     );


	    ybuf=YLINES+((2*64+2*VCOL_SPACE)*width); /* yllow is lowest */
	    if (!are_lines_empty(ybuf))
	    {
	       if (elinescount!=0)
	       {
		  paper_shift(elinescount);
		  elinescount=0;
	       }
	       find_horiz_ofsets(&left,&right,ybuf);
	       lxmsg(LXV_DEBUG,"Yellow offsets (pixels): left: %d, right %d\n",
		     left,right);
	       print_cols(left,right,128/2,192/2,ybuf,1);
	    }
	    elinescount+=16*2; // skip small space between 2 colors
	    /* now print magenta */
	    mbuf=MLINES+((1*64+VCOL_SPACE)*width); /* yllow is lowest */
	    if (!are_lines_empty(mbuf))
	    {
	       if (elinescount!=0)
	       {
		  paper_shift(elinescount);
		  elinescount=0;
	       }
	       find_horiz_ofsets(&left,&right,mbuf);
	       lxmsg(LXV_DEBUG,"Magenta offsets (pixels): left: %d, right %d\n",
		     left,right);
	       print_cols(left,right,64/2,128/2,mbuf,1);
	    }
	    elinescount+=16*2; // skip small space between 2 colors
	    /* now print cyan */
	    cbuf=CLINES; /* cyan is highest */
	    if (!are_lines_empty(cbuf))
	    {
	       if (elinescount!=0)
	       {
		  paper_shift(elinescount);
		  elinescount=0;
	       }
	       find_horiz_ofsets(&left,&right,cbuf);
	       lxmsg(LXV_DEBUG,"Cyan offsets (pixels): left: %d, right %d\n",
		     left,right);
	       print_cols(left,right,0,64/2,cbuf,1);
	    }

	    if (plates==3) /* CMY mode = all done */
	       elinescount+=32*2; // skip small space between 2 colors
	    else
	    {
	       /* K printing - not yet tested.... */
	       cbuf=KLINES+22*width; /* cyan is highest */
	       if (!are_lines_empty(cbuf))
	       {
		  if (elinescount!=0)
		  {
		     paper_shift(elinescount);
		     elinescount=0;
		  }
		  find_horiz_ofsets(&left,&right,cbuf);
		  lxmsg(LXV_DEBUG,"Black offsets (pixels): left: %d, right %d\n",
			left,right);
		  print_cols(left,right,0,64/2,cbuf,0);
	       }
	       elinescount+=32*2; // skip small space between 2 colors
	    }
	 } 
	 lremain-=plines;
      }
      paper_eject();
//            exit(-1);
      pagecount++;
      if (lines!=NULL)
      {
	 free(lines);
	 lines=NULL;
      }

   }

   lxmsg(LXV_DEBUG,"Two pass printing was used %d times to prevent"
	 "line overflow\n",twopasscount);
   lxmsg(LXV_STD,"printed (%d) page(s)\n",pagecount);
   closelog();
   return 0;
}
