/*
 * pnmraw2cmyk.c -- converts RBG image from PNM raw
 *                  to series of dithered CMY(K) PBMraw bitmaps
 *              
 * (C) 1999 Henryk Paluch, released under GNU GPL license, version 2
 * 
 * $Id: pnmraw2cmyk.c,v 1.6 1999/04/17 16:53:11 henryk Exp $
 */

#include<stdlib.h>
#include<stdio.h>
#include<errno.h>
#include<string.h>
#include<unistd.h> /* getopt() */
#include<syslog.h>
#include<ctype.h>

#include "dither.h"

#define DIV8(x) ( (x) >> 3 )
#define MOD8(x) ( (x) & 0x7 )

#ifndef MIN
#define MIN(x,y) ( (x) < (y) ? (x) : (y) )
#endif

int height;                 /* height of PBM bitmap (lines e.g., pixels) */
int width;                  /* width of line bitmap in pixels) */
int bwidth;                 /* real bitmap width (bytes) */
int obwidth;                /* output bitmap width (bytes) */
int obsize;		   /* output bitmap size (bytes) */
int plates;                /* number of color plates (3 or 4) */
unsigned char *lines=NULL; /* this is pixel buffer for one line */
unsigned char *olines=NULL;
static char *fprefix=NULL;
static int mayeof=0;
static int genk=0;
static int interlaced=0;  /* whether write 1 interlaced bitmap or
			   * more separated bitmaps
			   */

/*  parse_pnm_header() - decodes header of PNM bitmap
 *        sets global vars: height (pixels), 
 *                          bwidth (real bitamp width - bytes)
 *                          width  (memory bitmap width - bytes)
 *        PNMraw is simple RGB bitmap. It consists
 *        of text header and raw data bytes.
 *        Example: P6
 *                 # comment lines
 *                 width height
 *                 255
 *                 raw bytes
 *                 
 *                 width & height is in pixels,
 *                 every pixel is 3 byte value of RGB
 */


int parse_pnm_header(FILE *f)
{
   char buf[1024];
   int tmp;
   int retval=-1;
   char c;
   char *btypes[]={"monochromatic","grayscale","RGB"};
   
   if (fgets(buf,1023,f)==0  || buf[0]!='P')
   {
      if (mayeof)
	 return 0;
      else
      {
        fprintf(stderr,"Illegal pnmraw header"); 
	exit(-1);
      }
   }
   c=buf[1];
   if (!isdigit(c) || c<'4' || c>'6')
   {
      fprintf(stderr,"Unsupported bitmap type '%s'\n",buf);
      exit(-1);
   }
   c-='0';
   retval=c;
   do{
   fgets(buf,1023,f); /* skip comment line(s) */
   }while(buf[0]=='#');

   sscanf(buf,"%d %d\n",&width,&height);
   if (width <=0 || height <=0)
   {
      fprintf(stderr,"Illegal width (%d) or height (%d)\n",width,height);
      exit(-1);
   }
   if (c!=4) /* skip number of colors in case of P5 & P6 */
      fgets(buf,64,stdin); // scanf corrupts \n !!!
   if (c==6)
      bwidth= width*3; /* round width to whole byte */
   if (c==5)
      bwidth=width;
   if (c==4)
      bwidth=(width+7)/8;

   fprintf(stderr,"Detected %s RAW bitmap on input...\n",btypes[(int)(c-4)]);
   fprintf(stderr,"bwidth (bytes): %d (%d pixels), height (lines): %d\n",
	 bwidth,width, height);
   return retval;
}



/* read_lines() -- read 'readcount' data lines from PBM bitmap
 *                 to 'lines' -- bitmap in memory
 */

int read_line(FILE *f)
{
   int i,c;

   if (lines==NULL)
      lines=(char*)malloc(bwidth);
   /* zero the buffer */
/*   memset(lines,(char)0,bwidth); */
      c=fread(lines,1,bwidth,f);
      if (c==-1)
      {
	 perror("fread");
	 exit(-1);
      }
   return c;
}

static FILE *open_file(char *name)
{
   FILE *ff;

   if (name==NULL)
      return stdout;

   if ( (ff=fopen(name,"w+"))==NULL)
   {
      fprintf(stderr,"fopen '%s'",name);
      perror("");
      exit(-1);
   }
   return ff;
}

static void pbmgray2mono(char *fprefix)
{
   char buf[1024];
   unsigned char *mem,*omem;
   FILE *ff;
   int i,c,osize;

   if (fprefix!=NULL)
   {
      snprintf(buf,1023,"%s.pbm",fprefix);
      ff=open_file(buf);
   }
   else
      ff=stdout;
   fprintf(ff,"P4\n%d %d\n",width,height);
   mem=(unsigned char*)malloc(width);
   osize=(width+7)/8;
   omem=(unsigned char*)malloc(osize);
   for(i=0;i<height;i++)
   {
      if (fread(mem,1,width,stdin)!=bwidth)
      {
	 fprintf(stderr,"error while reading\n");
	 exit(-10);
      }
      dither_black(mem,i,width,width,omem);
      c=fwrite(omem,1,osize,ff);
      if (c!=osize)
      {
	 if (c>=0)
	    fprintf(stderr,"pbmgray2mono(): short write %d of %d\n",c,osize);
	 else
	    perror("pbmgray2mono(): fwrite");
	 exit(-1);
      }
      
   }

   free(omem);
   free(mem);
   if (ff!=stdout)
      fclose(ff);
}

static void copy_pbmraw(char *fprefix)
{
   char buf[1024];
   unsigned char *mem;
   FILE *ff;
   int c,cc;
   int rem;

   if (fprefix!=NULL)
   {
      snprintf(buf,1023,"%s.pbm",fprefix);
      ff=open_file(buf);
   }
   else
      ff=stdout;
   fprintf(ff,"P4\n%d %d\n",width,height);
   mem=(unsigned char*)malloc(bwidth);
   
   rem=height*bwidth;
   while(rem>0)
   {
//   fprintf(stderr,"Remainder %d\n",rem);
      c=fread(mem,1,MIN(rem,bwidth),stdin);
      if (c<=0)
      {
	 if (c==0)
	 {
	    fprintf(stderr,"Short read %d of %d bytes\n",c,MIN(rem,bwidth));
	 }
	 else
	 {
	    perror("pbmcopy: fread:");
	 }
	 exit(-1);
      }
	    
      rem-=c;
      cc=fwrite(mem,1,c,ff);
      if (cc!=c)
      {
	 if (cc>=0)
	 {
	    fprintf(stderr,"Short write %d of %d bytes\n",cc,bwidth);
	 }
	 else
	 {
	    perror("pbmcopy: fwrite:");
	 }
	 exit(-1);
      }
  }
   free(mem);
   if (ff!=stdout)
      fclose(ff);
}

static void write_pbm_bitmap(FILE *f,unsigned char *buf,char *filename,
      int plate)
{
   char *plnames[4]={"Cyan","Magenta","Yellow","Black"};
   int i;

   fprintf(f,"P4\n");
   fprintf(f,"# dithered separation '%s', %s plate\n",
		  filename,plnames[plate]);
   fprintf(f,"%d %d\n",width,height);
   for(i=0;i<height;i++)
   {
      if (fwrite(buf+i*obwidth,1,obwidth,f)!=obwidth)
      {
	 fprintf(stderr,"short write\n");
	 exit(-1);
      }
   }
}

static void usage(void)
{
   int i;
   fprintf(stderr,"pnmraw2cmyk PBMraw bitmap converter\n");
   fprintf(stderr,"   convert PNMraw RGB bitmap into separate"
	          " CMY(K) dithered PBMraw bitmaps\n");

   fprintf(stderr,"Usage: pnmraw2cmyk [ -h ] "
	          " < input.pnmraw > output.pbmraw (s)\n");
   fprintf(stderr,"       -h\tthis help text\n");
   fprintf(stderr,"       -o name\toutput to separate CMY(K) files"
	          " instead of stdout\n");
   fprintf(stderr,"       -k generate full CMYK instead of CMY\n");
   fprintf(stderr,"       -i write 1 interlaced bitmap rather than\n");
   fprintf(stderr,"          more separated bitmaps\n");
   exit(-1);
}

int main(int argc, char *argv[])
{
   int pagecount=0;
   int c,i;
   int bt;

   while( (c=getopt(argc, argv, ":o:hki"))!=-1 )
      switch(c){
	 case 'h':
	    usage();
	    break;
	 case 'o':
	    fprefix=optarg;
	    break;
	 case 'k':
	    genk=1;
	    break;
	 case 'i':
	    interlaced=1;
	    break;
	 default:
	    usage();
      }

   plates= (genk ? 4 : 3); /* number of color plates: 3 = CMY, 4 = CMYK */

   /* process raw PBM bitmap(s) */
   while((bt=parse_pnm_header(stdin))>0)
   {
      FILE *ff=NULL;
      int wrotehead=0;
      mayeof=1;

      if (bt==4)
      {
	 copy_pbmraw(fprefix);
	 continue;
      }
      if(bt==5)
      {
	 pbmgray2mono(fprefix);
	 continue;
      }
      obwidth=(width+7)/8;
      if (interlaced)
	 fprintf(stderr,"Creating interlaced bitmap...\n");
      fprintf(stderr,"Output bitmap width (%d bytes)\n",obwidth);
      if (interlaced)
	 obsize=obwidth;
      else
	 obsize=obwidth*height;
      fprintf(stderr,"Output buffer size per color (%d bytes)\n",obsize);

      olines=(unsigned char*)malloc(obsize*plates);
      fprintf(stderr,"Allocated %d bytes for %d bitmaps\n",obsize*plates,
	    plates);
      /* now it is time to dither */
      for(i=0;i<height;i++)
      {
	 int memofs;
	 if (interlaced)
	    memofs=0; /* non interlaced are whole stored in memory */
	 else
	    memofs=obwidth*i; /* interlaced out - only current line in mem */
	 if (read_line(stdin)!=bwidth)
	 {
	    fprintf(stderr,"short read\n");
	    exit(-1);
	 }
	 dither_cmyk(lines, /* input row */
	       i,      /* row number */
	       width,  /* width of source in pixels */
	       width,  /* width of dest in pixels */
	       olines+0*obsize+memofs, /* cyan outbuf */
	       olines+1*obsize+memofs, /* magenta outbuf */
	       olines+2*obsize+memofs, /* yellow outbuf */
	       (genk ?                    /* black outbuf */
		olines+3*obsize+memofs
		: NULL ));

	 if (interlaced)
	 {
	    if (!wrotehead)
	    {
	       wrotehead=1;
	       if (fprefix!=NULL)
	       {
		  char buf[1024];
		  snprintf(buf,1023,"%si-%03d.pbm",
			fprefix, pagecount);
		  ff=open_file(buf);
	       }
	       else
	       {
		  ff=stdout;
	       }
	       /* write interlaced pbm header now */
	       fprintf(ff,"P4\n#PI%d %d %d\n%d %d\n",plates,
		     width,height,width,height*plates);

	    }
	    if ( fwrite(olines,1,obsize*plates,ff)!=obsize*plates)
	    {
	       perror("fwrite");
	       exit(-1);
	    }
	 }
      }
      if (!interlaced) 
      {
	 if (fprefix!=NULL)
	 {
	    char pnames[]="cmyk";
	    int y;
	    char buf[1024];
	    for(y=0;y<plates;y++)
	    {
	       snprintf(buf,1023,"%s%c-%03d.pbm",fprefix,pnames[y],pagecount);
	       ff=open_file(buf);
	       write_pbm_bitmap(ff,olines+y*obsize,buf,y);
	       fclose(ff);
	       ff=NULL;
	    }
	 }
	 else
	 {
	    int y;
	    for(y=0;y<plates;y++)
	       write_pbm_bitmap(stdout,olines+y*obsize,"(stdout)",y);
	 }
      }
      pagecount++;
      if (lines!=NULL)
      {
	 free(lines);
	 lines=NULL;
      }
      if (olines!=NULL)
      {
	 free(olines);
	 olines=NULL;
      }
      if (ff!=NULL && ff!=stdout)
      {
	 fclose(ff);
	 ff=NULL;
      }
   }

   return 0;
}
